API_STAGE = "dev"
APP_FUNCTION = "raises_exception"
APP_MODULE = "tests.test_handler"
BINARY_SUPPORT = False
CONTEXT_HEADER_MAPPINGS = {}
DEBUG = "True"
DJANGO_SETTINGS = None
DOMAIN = "api.example.com"
ENVIRONMENT_VARIABLES = {}
LOG_LEVEL = "DEBUG"
PROJECT_NAME = "raises_exception"
COGNITO_TRIGGER_MAPPING = {}
EXCEPTION_HANDLER = "tests.test_handler.mocked_exception_handler"
