API_STAGE = "dev"
APP_FUNCTION = "handler_for_events"
APP_MODULE = "tests.test_event_script_app"
DEBUG = "True"
DJANGO_SETTINGS = None
DOMAIN = "api.example.com"
ENVIRONMENT_VARIABLES = {}
LOG_LEVEL = "DEBUG"
PROJECT_NAME = "test_event_script_app"
COGNITO_TRIGGER_MAPPING = {}
